# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
import click
import six

from databricks_cli.click_types import ContextObject
from databricks_cli.configure.provider import get_config, \
    update_and_persist_config, ProfileConfigProvider
from databricks_cli.oauth.oauth import check_and_refresh_access_token
from databricks_cli.utils import InvalidConfigurationError
from databricks_cli.sdk import ApiClient
from databricks_cli.sdk.version import API_VERSIONS


def provide_api_client(function):
    """
    Injects the api_client keyword argument to the wrapped function.
    All callbacks wrapped by provide_api_client expect the argument ``profile`` to be passed in.
    """
    @six.wraps(function)
    def decorator(*args, **kwargs):
        ctx = click.get_current_context()
        command_name = "-".join(ctx.command_path.split(" ")[1:])
        command_name += "-" + str(uuid.uuid1())
        profile = get_profile_from_context()
        if profile:
            # If we request a specific profile, only get credentials from there.
            config = ProfileConfigProvider(profile).get_config()
        else:
            # If unspecified, use the default provider, or allow for user overrides.
            config = get_config()
        if not config or not config.is_valid:
            raise InvalidConfigurationError.for_profile(profile)

        # This checks if an OAuth access token has expired and will attempt to refresh it if
        # a refresh token is present
        if config.host and config.token and config.refresh_token:
            config.token, config.refresh_token, updated = \
                check_and_refresh_access_token(config.host, config.token, config.refresh_token)
            if updated:
                update_and_persist_config(profile, config)

        kwargs['api_client'] = _get_api_client(config, command_name)

        return function(*args, **kwargs)
    decorator.__doc__ = function.__doc__
    return decorator


def get_profile_from_context():
    ctx = click.get_current_context()
    context_object = ctx.ensure_object(ContextObject)
    return context_object.get_profile()


def debug_option(f):
    def callback(ctx, param, value):  # NOQA
        context_object = ctx.ensure_object(ContextObject)
        context_object.set_debug(value)
    return click.option('--debug', is_flag=True, callback=callback,
                        expose_value=False, help="Debug Mode. Shows full stack trace on error.")(f)


def profile_option(f):
    def callback(ctx, param, value):  # NOQA
        if value is not None:
            context_object = ctx.ensure_object(ContextObject)
            context_object.set_profile(value)
    return click.option('--profile', required=False, default=None, callback=callback,
                        expose_value=False,
                        help='CLI connection profile to use. The default profile is "DEFAULT".')(f)


def api_version_option(f):
    return click.option('--version', required=False, default=None, type=click.Choice(API_VERSIONS),
                        help='Override the API version used to call databricks.')(f)


def _get_api_client(config, command_name=""):
    verify = config.insecure is None
    if config.is_valid_with_token:
        return ApiClient(host=config.host, token=config.token, verify=verify,
                         command_name=command_name, jobs_api_version=config.jobs_api_version)
    return ApiClient(user=config.username, password=config.password,
                     host=config.host, verify=verify, command_name=command_name,
                     jobs_api_version=config.jobs_api_version)



"""Implement Databricks Groups API, interfacing with the GroupsService."""
# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from databricks_cli.sdk import GroupsService


class GroupsApi(object):
    """Implement the databricks '2.0/groups' API Interface."""

    def __init__(self, api_client):
        self.client = GroupsService(api_client)

    def add_member(self, parent_name, user_name, group_name):
        """
        Only one of ``user_name`` or ``group_name`` should be provided.
        """
        assert bool(user_name is not None) ^ bool(group_name is not None)
        return self.client.add_to_group(parent_name=parent_name,
                                        user_name=user_name,
                                        group_name=group_name)

    def create(self, group_name):
        """Create a new group with the given name."""
        return self.client.create_group(group_name)

    def list_members(self, group_name):
        """Return all of the members of a particular group."""
        return self.client.get_group_members(group_name)

    def list_all(self):
        """Return all of the groups in an organization."""
        return self.client.get_groups()

    def list_parents(self, user_name, group_name):
        """
        Only one of ``user_name`` or ``group_name`` should be provided.

        Retrieve all groups in which a given user or group is a member.

        Note: this method is non-recursive - it will return all groups in
        which the given user or group is a member but not the groups in which
        those groups are members).
        """
        assert bool(user_name is not None) ^ bool(group_name is not None)
        return self.client.get_groups_for_principal(user_name=user_name, group_name=group_name)

    def remove_member(self, parent_name, user_name, group_name):
        """
        Only one of ``user_name`` or ``group_name`` should be provided.
        """
        assert bool(user_name is not None) ^ bool(group_name is not None)
        return self.client.remove_from_group(parent_name=parent_name,
                                             user_name=user_name,
                                             group_name=group_name)

    def delete(self, group_name):
        """Remove a group from this organization."""
        return self.client.remove_group(group_name)



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from databricks_cli.configure.config import profile_option, debug_option
from databricks_cli.libraries.cli import libraries_group
from databricks_cli.version import print_version_callback, version
from databricks_cli.utils import CONTEXT_SETTINGS
from databricks_cli.configure.cli import configure_cli
from databricks_cli.dbfs.cli import dbfs_group
from databricks_cli.workspace.cli import workspace_group
from databricks_cli.jobs.cli import jobs_group
from databricks_cli.clusters.cli import clusters_group
from databricks_cli.cluster_policies.cli import cluster_policies_group
from databricks_cli.runs.cli import runs_group
from databricks_cli.secrets.cli import secrets_group
from databricks_cli.stack.cli import stack_group
from databricks_cli.groups.cli import groups_group
from databricks_cli.tokens.cli import tokens_group
from databricks_cli.instance_pools.cli import instance_pools_group
from databricks_cli.pipelines.cli import pipelines_group
from databricks_cli.repos.cli import repos_group
from databricks_cli.unity_catalog.cli import unity_catalog_group


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('--version', '-v', is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=version)
@debug_option
@profile_option
def cli():
    pass


cli.add_command(configure_cli, name='configure')
cli.add_command(dbfs_group, name='fs')
cli.add_command(workspace_group, name='workspace')
cli.add_command(jobs_group, name='jobs')
cli.add_command(clusters_group, name='clusters')
cli.add_command(cluster_policies_group, name='cluster-policies')
cli.add_command(runs_group, name='runs')
cli.add_command(libraries_group, name='libraries')
cli.add_command(secrets_group, name='secrets')
cli.add_command(stack_group, name='stack')
cli.add_command(groups_group, name='groups')
cli.add_command(tokens_group, name='tokens')
cli.add_command(instance_pools_group, name="instance-pools")
cli.add_command(pipelines_group, name='pipelines')
cli.add_command(repos_group, name='repos')
cli.add_command(unity_catalog_group, name='unity-catalog')

if __name__ == "__main__":
    cli()



# Databricks CLI
# Copyright 2022 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from databricks_cli.click_types import JsonClickType
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.unity_catalog.api import UnityCatalogApi
from databricks_cli.unity_catalog.utils import hide, json_file_help, json_string_help, \
    mc_pretty_format
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, json_cli_base


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Create a new schema.')
@click.option('--catalog-name', required=True, help='Parent catalog of new schema.')
@click.option('--name', required=True,
              help='Name of new schema, relative to parent catalog.')
@click.option('--comment', default=None, required=False,
              help='Free-form text description.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def create_schema_cli(api_client, catalog_name, name, comment):
    """
    Create a new schema in the specified catalog.
    """
    schema_json = UnityCatalogApi(api_client).create_schema(catalog_name, name, comment)
    click.echo(mc_pretty_format(schema_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List schemas.')
@click.option('--catalog-name', required=True,
              help='Name of the parent catalog for schemas of interest.')
@click.option('--name-pattern', default=None,
              help='SQL LIKE pattern that the schema name must match to be in list.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_schemas_cli(api_client, catalog_name, name_pattern):
    """
    List schemas.
    """
    schemas_json = UnityCatalogApi(api_client).list_schemas(catalog_name, name_pattern)
    click.echo(mc_pretty_format(schemas_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Get a schema.')
@click.option('--full-name', required=True,
              help='Full name (<catalog>.<schema>) of the schema to get.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_schema_cli(api_client, full_name):
    """
    Get a schema.
    """
    schema_json = UnityCatalogApi(api_client).get_schema(full_name)
    click.echo(mc_pretty_format(schema_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Update a schema.')
@click.option('--full-name', required=True,
              help='Full name (<catalog>.<schema>) of the schema to update.')
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='PATCH', path='/schemas/{name}'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='PATCH', path='/schemas/{name}'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def update_schema_cli(api_client, full_name, json_file, json):
    """
    Update a schema.

    The public specification for the JSON request is in development.
    """
    json_cli_base(json_file, json,
                  lambda json: UnityCatalogApi(api_client).update_schema(full_name, json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Delete a schema.')
@click.option('--full-name', required=True,
              help='Full name (<catalog>.<schema>) of the schema to delete.')
@click.option('--purge', '-p', is_flag=True, default=False,
              help='Purge all child schemas and tables of catalog.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_schema_cli(api_client, full_name, purge):
    """
    Delete a schema.
    """
    if purge:
        (catalog_name, schema_name) = full_name.split('.')
        click.echo("Purging all tables from schema %s in catalog %s" % (schema_name, catalog_name))

        tables_response = UnityCatalogApi(api_client).list_tables(catalog_name, schema_name, None)
        for t in tables_response['tables']:
            click.echo("Deleting table: %s" % (t['full_name']))
            UnityCatalogApi(api_client).delete_table(t['full_name'])

    UnityCatalogApi(api_client).delete_schema(full_name)


@click.group()
def schemas_group():  # pragma: no cover
    pass


def register_schema_commands(cmd_group):
    # Register deprecated "verb-noun" commands for backward compatibility.
    cmd_group.add_command(hide(create_schema_cli), name='create-schema')
    cmd_group.add_command(hide(list_schemas_cli), name='list-schemas')
    cmd_group.add_command(hide(get_schema_cli), name='get-schema')
    cmd_group.add_command(hide(update_schema_cli), name='update-schema')
    cmd_group.add_command(hide(delete_schema_cli), name='delete-schema')

    # Register command group.
    schemas_group.add_command(create_schema_cli, name='create')
    schemas_group.add_command(list_schemas_cli, name='list')
    schemas_group.add_command(get_schema_cli, name='get')
    schemas_group.add_command(update_schema_cli, name='update')
    schemas_group.add_command(delete_schema_cli, name='delete')
    cmd_group.add_command(schemas_group, name='schemas')



# Databricks CLI
# Copyright 2022 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

import click

from databricks_cli.click_types import JsonClickType
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.unity_catalog.api import UnityCatalogApi
from databricks_cli.unity_catalog.utils import hide, json_file_help, json_string_help, \
    mc_pretty_format
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, json_cli_base


#############  Storage Credential Commands  ############

def fill_credential(
    data, aws_iam_role_arn, az_sp_directory_id, az_sp_application_id,
    az_sp_client_secret, az_mi_access_connector_id, az_mi_id, gcp_sak_email,
    gcp_sak_private_key_id, gcp_sak_private_key):
    if aws_iam_role_arn is not None:
        data['aws_iam_role'] = {
            'role_arn': aws_iam_role_arn
        }

    if ((az_sp_directory_id is not None) or (az_sp_application_id is not None) or
        (az_sp_client_secret is not None)):
        data['azure_service_principal'] = {
            'directory_id': az_sp_directory_id,
            'application_id': az_sp_application_id,
            'client_secret': az_sp_client_secret
        }

    if (az_mi_access_connector_id is not None) or (az_mi_id is not None):
        data['azure_managed_identity'] = {
            'access_connector_id': az_mi_access_connector_id,
            'managed_identity_id': az_mi_id
        }

    if ((gcp_sak_email is not None) or (gcp_sak_private_key_id is not None) or 
        (gcp_sak_private_key is not None)):
        data['gcp_service_account_key'] = {
            'email': gcp_sak_email,
            'private_key_id': gcp_sak_private_key_id,
            'private_key': gcp_sak_private_key
        }


def create_update_common_options(f):
    @click.option('--aws-iam-role-arn', default=None,
                help='The Amazon Resource Name (ARN) of the AWS IAM role for S3 data access.')
    @click.option('--az-sp-directory-id', default=None,
                help=(
                    'The directory ID corresponding to the Azure Active Directory (AAD) '
                    'tenant of the application.'))
    @click.option('--az-sp-application-id', default=None,
                help=(
                    'The application ID of the application registration within the referenced '
                    'AAD tenant.'))
    @click.option('--az-sp-client-secret', default=None,
                help='The client secret generated for the above app ID in AAD.')
    @click.option('--az-mi-access-connector-id', default=None,
                help=(
                    'The Azure resource ID of the Azure Databricks Access Connector. '
                    'Use the format, '
                    '/subscriptions/{guid}/resourceGroups/{rg-name}/providers/Microsoft.Databricks'
                    '/accessConnectors/{connector-name} .'))
    @click.option('--az-mi-id', default=None,
                help=(
                    'The Azure resource ID of the managed identity. Use the format, '
                    '/subscriptions/{guid}/resourceGroups/{rg-name}/providers'
                    '/Microsoft.ManagedIdentity/userAssignedIdentities/{identity-name} .'
                    'This is only available for user-assigned identities. '
                    'For system-assigned identities, access-connector-id is used to identify '
                    'the identity. If this flag is not provided, '
                    'then we assume that it is using the system-assigned identity.'))
    @click.option('--gcp-sak-email', default=None,
                help=(
                    'Credential for GCP Service Account Key. '
                    'The email of the service account.'))
    @click.option('--gcp-sak-private-key-id', default=None,
                help=(
                    'Credential for GCP Service Account Key. '
                    'The ID of the service account\'s private key.'))
    @click.option('--gcp-sak-private-key', default=None,
                help=(
                    'Credential for GCP Service Account Key. '
                    'The service account\'s RSA private key.'))
    @click.option('--comment', default=None,
                help='Free-form text description.')
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        f(*args, **kwargs)
    return wrapper


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Create storage credential.')
@click.option('--name', default=None,
              help='Name of new storage credential')
@create_update_common_options
@click.option('--skip-validation', '-s', 'skip_val', is_flag=True, default=False,
              help='Skip the validation of new credential info before creation')
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='POST', path='/storage-credentials'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='POST', path='/storage-credentials'))
@debug_option
@profile_option
# UC's createStorageCredential returns a 401 when validation fails; that translates to
# a misleading error when eat_exceptions is enabled:
#   Your authentication information may be incorrect. Please reconfigure with ``dbfs configure``
# Until that is fixed (should return a 400), show full error trace.
#@eat_exceptions
@provide_api_client
def create_credential_cli(api_client, name, aws_iam_role_arn,
                          az_sp_directory_id, az_sp_application_id, az_sp_client_secret,
                          az_mi_access_connector_id, az_mi_id, gcp_sak_email,
                          gcp_sak_private_key_id, gcp_sak_private_key, comment,
                          skip_val, json_file, json):
    """
    Create new storage credential.

    The public specification for the JSON request is in development.
    """
    has_credential_flag = (
        (aws_iam_role_arn is not None) or
        (az_sp_directory_id is not None) or (az_sp_application_id is not None) or
        (az_sp_client_secret is not None) or (az_mi_access_connector_id is not None) or
        (az_mi_id is not None) or (gcp_sak_email is not None) or
        (gcp_sak_private_key_id is not None) or (gcp_sak_private_key is not None))
    if ((name is not None) or has_credential_flag or (comment is not None)):
        if (json_file is not None) or (json is not None):
            raise ValueError('Cannot specify JSON if any other creation flags are specified')
        data = {
            'name': name,
            'comment': comment
        }

        fill_credential(
            data, aws_iam_role_arn, az_sp_directory_id, az_sp_application_id,
            az_sp_client_secret, az_mi_access_connector_id, az_mi_id, gcp_sak_email,
            gcp_sak_private_key_id, gcp_sak_private_key)

        cred_json = UnityCatalogApi(api_client).create_storage_credential(data, skip_val)
        click.echo(mc_pretty_format(cred_json))
    else:
        json_cli_base(json_file, json,
                      lambda json: UnityCatalogApi(api_client).create_storage_credential(json,
                                                                                         skip_val),
                      encode_utf8=True)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List storage credentials.')
@click.option('--name-pattern', default=None,
              help='SQL LIKE pattern that the credential name must match to be in list.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_credentials_cli(api_client, name_pattern):
    """
    List storage credentials.
    """
    creds_json = UnityCatalogApi(api_client).list_storage_credentials(name_pattern)
    click.echo(mc_pretty_format(creds_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Get a storage credential.')
@click.option('--name', required=True,
              help='Name of the storage credential to get.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_credential_cli(api_client, name):
    """
    Get a storage credential.
    """
    cred_json = UnityCatalogApi(api_client).get_storage_credential(name)
    click.echo(mc_pretty_format(cred_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Update a storage credential.')
@click.option('--name', required=True,
              help='Name of the storage credential to update.')
@click.option('--new-name', default=None, help='New name of the storage credential.')
@create_update_common_options
@click.option('--owner', default=None,
              help='Owner of the storage credential.')
@click.option('--skip-validation', '-s', 'skip_val', is_flag=True, default=False,
              help='Skip the validation of new credential info before update')
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='PATCH', path='/storage-credentials/{name}'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='PATCH', path='/storage-credentials/{name}'))
@debug_option
@profile_option
# See comment for create-storage-credential
#@eat_exceptions
@provide_api_client
def update_credential_cli(api_client, name, new_name, aws_iam_role_arn,
                          az_sp_directory_id, az_sp_application_id, az_sp_client_secret,
                          az_mi_access_connector_id, az_mi_id, gcp_sak_email,
                          gcp_sak_private_key_id, gcp_sak_private_key, comment, owner,
                          skip_val, json_file, json):
    """
    Update a storage credential.

    The public specification for the JSON request is in development.
    """
    has_credential_flag = (
        (aws_iam_role_arn is not None) or
        (az_sp_directory_id is not None) or (az_sp_application_id is not None) or
        (az_sp_client_secret is not None) or (az_mi_access_connector_id is not None) or
        (az_mi_id is not None) or (gcp_sak_email is not None) or
        (gcp_sak_private_key_id is not None) or (gcp_sak_private_key is not None))
    if ((new_name is not None) or has_credential_flag or
        (comment is not None) or (owner is not None)):
        if (json_file is not None) or (json is not None):
            raise ValueError('Cannot specify JSON if any other update flags are specified')
        data = {
            'name': new_name,
            'comment': comment,
            'owner': owner
        }

        fill_credential(
            data, aws_iam_role_arn, az_sp_directory_id, az_sp_application_id,
            az_sp_client_secret, az_mi_access_connector_id, az_mi_id, gcp_sak_email,
            gcp_sak_private_key_id, gcp_sak_private_key)

        cred_json = UnityCatalogApi(api_client).update_storage_credential(name, data, skip_val)
        click.echo(mc_pretty_format(cred_json))
    else:
        json_cli_base(json_file, json,
                     lambda json: UnityCatalogApi(api_client).update_storage_credential(name,
                                                                                        json,
                                                                                        skip_val),
                     encode_utf8=True)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Delete a storage credential.')
@click.option('--name', required=True,
              help='Name of the storage credential to delete.')
@click.option('--force', '-f', is_flag=True, default=False,
              help='Force deletion even if credential has dependent tables/locations')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_credential_cli(api_client, name, force):
    """
    Delete a storage credential.
    """
    UnityCatalogApi(api_client).delete_storage_credential(name, force)


@click.group()
def storage_credentials_group():  # pragma: no cover
    pass


def register_cred_commands(cmd_group):
    # Register deprecated "verb-noun" commands for backward compatibility.
    cmd_group.add_command(hide(create_credential_cli), name='create-storage-credential')
    cmd_group.add_command(hide(list_credentials_cli), name='list-storage-credentials')
    cmd_group.add_command(hide(get_credential_cli), name='get-storage-credential')
    cmd_group.add_command(hide(update_credential_cli), name='update-storage-credential')
    cmd_group.add_command(hide(delete_credential_cli), name='delete-storage-credential')

    # Register command group.
    storage_credentials_group.add_command(create_credential_cli, name='create')
    storage_credentials_group.add_command(list_credentials_cli, name='list')
    storage_credentials_group.add_command(get_credential_cli, name='get')
    storage_credentials_group.add_command(update_credential_cli, name='update')
    storage_credentials_group.add_command(delete_credential_cli, name='delete')
    cmd_group.add_command(storage_credentials_group, name='storage-credentials')



# Databricks CLI
# Copyright 2022 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from databricks_cli.click_types import JsonClickType, OneOfOption
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.unity_catalog.api import UnityCatalogApi
from databricks_cli.unity_catalog.utils import hide, json_file_help, json_string_help, \
    mc_pretty_format
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, json_cli_base


PERMISSIONS_OBJ_TYPES = [
    'metastore', 'catalog', 'schema', 'table', 'storage-credential', 'external-location'
]


def _get_perm_securable_name_and_type(metastore_id, catalog_name, schema_full_name,
                                      table_full_name, credential_name, location_name):
    if metastore_id:
        return ('metastore', metastore_id)
    elif catalog_name:
        return ('catalog', catalog_name)
    elif schema_full_name:
        return ('schema', schema_full_name)
    elif table_full_name:
        return ('table', table_full_name)
    elif credential_name:
        return ('storage-credential', credential_name)
    else:
        return ('external-location', location_name)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Get permissions on a securable.')
@click.option('--metastore', cls=OneOfOption, default=None,
              one_of=PERMISSIONS_OBJ_TYPES,
              help='ID of metastore of interest')
@click.option('--catalog', cls=OneOfOption, default=None,
              one_of=PERMISSIONS_OBJ_TYPES,
              help='Name of catalog of interest')
@click.option('--schema', cls=OneOfOption, default=None,
              one_of=PERMISSIONS_OBJ_TYPES,
              help='Full name of schema of interest')
@click.option('--table', cls=OneOfOption, default=None,
              one_of=PERMISSIONS_OBJ_TYPES,
              help='Full name of table of interest')
@click.option('--storage-credential', cls=OneOfOption, default=None,
              one_of=PERMISSIONS_OBJ_TYPES,
              help='Name of the storage credential of interest')
@click.option('--external-location', cls=OneOfOption, default=None,
              one_of=PERMISSIONS_OBJ_TYPES,
              help='Name of the external location of interest')
@click.option('--effective', is_flag=True, default=False,
              help='Get effective permissions (including inherited privileges)')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_permissions_cli(api_client, metastore, catalog, schema, table, storage_credential,
                        external_location, effective):
    """
    Get permissions on a securable.
    """
    sec_type, sec_name = _get_perm_securable_name_and_type(metastore, catalog, schema, table,
                                                           storage_credential, external_location)

    if effective:
        perm_json = UnityCatalogApi(api_client).get_effective_permissions(sec_type, sec_name)
    else:
        perm_json = UnityCatalogApi(api_client).get_permissions(sec_type, sec_name)
    click.echo(mc_pretty_format(perm_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='update permissions on a securable.')
@click.option('--metastore', cls=OneOfOption, default=None,
              one_of=PERMISSIONS_OBJ_TYPES,
              help='ID of metastore of interest')
@click.option('--catalog', cls=OneOfOption, default=None,
              one_of=PERMISSIONS_OBJ_TYPES,
              help='Name of catalog of interest')
@click.option('--schema', cls=OneOfOption, default=None,
              one_of=PERMISSIONS_OBJ_TYPES,
              help='Full name of schema of interest')
@click.option('--table', cls=OneOfOption, default=None,
              one_of=PERMISSIONS_OBJ_TYPES,
              help='Full name of table of interest')
@click.option('--storage-credential', cls=OneOfOption, default=None,
              one_of=PERMISSIONS_OBJ_TYPES,
              help='Name of the storage credential of interest')
@click.option('--external-location', cls=OneOfOption, default=None,
              one_of=PERMISSIONS_OBJ_TYPES,
              help='Name of the external location of interest')
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='PATCH', path='/permissions/{securable}/{id}'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='PATCH', path='/permissions/{securable}/{id}'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def update_permissions_cli(api_client, metastore, catalog, schema, table, storage_credential,
                           external_location, json_file, json):
    """
    Update permissions on a securable.

    The public specification for the JSON request is in development.
    """
    sec_type, sec_name = _get_perm_securable_name_and_type(metastore, catalog, schema, table,
                                                           storage_credential, external_location)

    json_cli_base(json_file, json,
                  lambda json: UnityCatalogApi(api_client).update_permissions(sec_type, sec_name,
                                                                              json),
                  encode_utf8=True)


@click.group()
def permissions_group():  # pragma: no cover
    pass


def register_perms_commands(cmd_group):
    # Register deprecated "verb-noun" commands for backward compatibility.
    cmd_group.add_command(hide(get_permissions_cli), name='get-permissions')
    cmd_group.add_command(hide(update_permissions_cli), name='update-permissions')

    # Register command group.
    permissions_group.add_command(get_permissions_cli, name='get')
    permissions_group.add_command(update_permissions_cli, name='update')
    cmd_group.add_command(permissions_group, name='permissions')



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hashlib import sha1
import os
import copy

from six.moves import urllib

from databricks_cli.sdk import DeltaPipelinesService
from databricks_cli.dbfs.api import DbfsApi
from databricks_cli.dbfs.dbfs_path import DbfsPath

BUFFER_SIZE = 1024 * 64
base_pipelines_dir = 'dbfs:/pipelines/code'


class PipelinesApi(object):
    def __init__(self, api_client):
        self.client = DeltaPipelinesService(api_client)
        self.dbfs_client = DbfsApi(api_client)

    def create(self, settings, settings_dir, allow_duplicate_names, headers=None):
        data = self._upload_libraries_and_update_settings(settings, settings_dir)
        data['allow_duplicate_names'] = allow_duplicate_names
        return self.client.client.perform_query('POST', '/pipelines', data=data,
                                                headers=headers)

    def edit(self, settings, settings_dir, allow_duplicate_names, headers=None):
        data = self._upload_libraries_and_update_settings(settings, settings_dir)
        data['allow_duplicate_names'] = allow_duplicate_names
        pipeline_id = data['id']
        self.client.client.perform_query('PUT', '/pipelines/{}'.format(pipeline_id), data=data,
                                         headers=headers)

    def delete(self, pipeline_id, headers=None):
        self.client.delete(pipeline_id, headers)

    def get(self, pipeline_id, headers=None):
        return self.client.get(pipeline_id, headers)

    def list(self, headers=None):
        def call(page_token=None, max_results=None, order_by=None):
            _data = {}
            if page_token:
                _data["page_token"] = page_token
            if max_results:
                _data["max_results"] = max_results
            if order_by:
                _data["order_by"] = order_by

            return self.client.client.perform_query(
                'GET', '/pipelines', data=_data, headers=headers)

        response = call()
        pipelines = response.get("statuses", [])

        while "next_page_token" in response:
            response = call(page_token=response["next_page_token"])
            pipelines.extend(response.get("statuses", []))
        return pipelines

    def start_update(self, pipeline_id, full_refresh=None, headers=None):
        return self.client.start_update(pipeline_id, full_refresh=full_refresh, headers=headers)

    def stop(self, pipeline_id, headers=None):
        self.client.stop(pipeline_id, headers)

    def _upload_libraries_and_update_settings(self, settings, settings_dir):
        settings = copy.deepcopy(settings)
        lib_objects = LibraryObject.from_json(settings.get('libraries', []))
        local_lib_objects, external_lib_objects = self._identify_local_libraries(lib_objects)

        settings['libraries'] = LibraryObject.to_json(
            external_lib_objects + self._upload_local_libraries(
                settings_dir, local_lib_objects))
        return settings

    @staticmethod
    def _identify_local_libraries(lib_objects):
        """
        Partitions the given set of libraries into local libraries i.e. libraries that should
        be uploaded to DBFS, and non-local libraries. Jars or whls with a file scheme or no scheme
        at all are (currently) considered local libraries.

        :param lib_objects: List[LibraryObject]
        :return: List[List[LibraryObject], List[LibraryObject]] ([Local, External])
        """
        local_lib_objects, external_lib_objects = [], []
        for lib_object in lib_objects:
            if lib_object.lib_type not in ['jar', 'whl']:
                external_lib_objects.append(lib_object)
                continue

            parsed_uri = urllib.parse.urlparse(lib_object.path)
            if parsed_uri.scheme == '':
                local_lib_objects.append(lib_object)
            elif parsed_uri.scheme.lower() == 'file':
                # exactly 1 or 3
                if parsed_uri.path.startswith('//') or parsed_uri.netloc != '':
                    raise RuntimeError('invalid file uri scheme, '
                                       'did you mean to use file:/ or file:///')
                local_lib_objects.append(LibraryObject(lib_object.lib_type, parsed_uri.path))
            else:
                external_lib_objects.append(lib_object)
        return local_lib_objects, external_lib_objects

    def _upload_local_libraries(self, settings_dir, local_lib_objects):
        relative_local_lib_objects = [
            LibraryObject(
                llo.lib_type, os.path.join(settings_dir, llo.path)) for llo in local_lib_objects]
        remote_lib_objects = [LibraryObject(rllo.lib_type, self._get_hashed_path(rllo.path))
                              for rllo in relative_local_lib_objects]
        transformed_remote_lib_objects = [LibraryObject(rlo.lib_type, DbfsPath(rlo.path))
                                          for rlo in remote_lib_objects]
        upload_files = [llo_tuple for llo_tuple in
                        zip(relative_local_lib_objects, transformed_remote_lib_objects)
                        if not self.dbfs_client.file_exists(llo_tuple[1].path)]

        for llo, rlo in upload_files:
            self.dbfs_client.put_file(llo.path, rlo.path, False)

        return remote_lib_objects

    @staticmethod
    def _get_hashed_path(path):
        """
        Finds the corresponding dbfs file path for the file located at the supplied path by
        calculating its hash using SHA1.
        :param path: Local File Path
        :return: Remote Path (pipeline_base_dir + file_hash (dot) file_extension)
        """
        hash_buffer = sha1()
        with open(path, 'rb') as f:
            while True:
                data = f.read(BUFFER_SIZE)
                if not data:
                    break
                hash_buffer.update(data)

        file_hash = hash_buffer.hexdigest()
        # splitext includes the period in the extension
        extension = os.path.splitext(path)[1][1:]
        if extension == 'whl':
            # Wheels need to follow the format described in the PEP, so we simply
            # pre-pend the content hash to the wheel_name
            # basename in Python returns the extension as well
            wheel_name = os.path.basename(path)
            path = '{}/{}/{}'.format(base_pipelines_dir, file_hash, wheel_name)
        else:
            path = '{}/{}.{}'.format(base_pipelines_dir, file_hash, extension)
        return path


class LibraryObject(object):
    def __init__(self, lib_type, lib_path):
        self.path = lib_path
        self.lib_type = lib_type

    @classmethod
    def from_json(cls, libraries):
        """
        Serialize Libraries into LibraryObjects
        :param libraries: List[Dictionary{String, String}]
        :return: List[LibraryObject]
        """
        lib_objects = []
        for library in libraries:
            for lib_type, path in library.items():
                lib_objects.append(LibraryObject(lib_type, path))
        return lib_objects

    @classmethod
    def to_json(cls, lib_objects):
        """
        Deserialize LibraryObjects
        :param lib_objects: List[LibraryObject]
        :return: List[Dictionary{String, String}]
        """
        libraries = []
        for lib_object in lib_objects:
            libraries.append({lib_object.lib_type: lib_object.path})
        return libraries

    def __eq__(self, other):
        if not isinstance(other, LibraryObject):
            return NotImplemented
        return self.path == other.path and self.lib_type == other.lib_type



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import requests

try:
    from urlparse import urlparse, urljoin
except ImportError:
    from urllib.parse import urlparse, urljoin

import click

from databricks_cli.click_types import PipelineSpecClickType, \
    PipelineSettingClickType, PipelineIdClickType
from databricks_cli.version import print_version_callback, version
from databricks_cli.pipelines.api import PipelinesApi
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.utils import pipelines_exception_eater, CONTEXT_SETTINGS, pretty_format, \
    error_and_quit

try:
    json_parse_exception = json.decoder.JSONDecodeError
except AttributeError:  # Python 2
    json_parse_exception = ValueError


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Creates a pipeline.')
@click.argument('settings_arg', default=None, required=False)
@click.option('--settings', default=None,
              type=PipelineSettingClickType(), help=PipelineSettingClickType.help)
@click.option('--allow-duplicate-names', is_flag=True,
              help="If true, skips duplicate name checking while creating the pipeline.")
@debug_option
@profile_option
@pipelines_exception_eater
@provide_api_client
def create_cli(api_client, settings_arg, settings, allow_duplicate_names):
    # pylint: disable=line-too-long
    """
    Creates a pipeline specified by the pipeline settings. The pipeline settings are a
    JSON document that defines a Delta Live Tables pipeline on Databricks.

    To use a file containing the pipeline settings, pass the file path to the command as
    an argument or with the --settings option. If the pipeline creation is successful, logs
    the URL and the ID of the new pipeline to STDOUT.

    Specification for the pipeline settings JSON can be found at
    https://docs.databricks.com/data-engineering/delta-live-tables/delta-live-tables-configuration.html

    If a pipeline with the same name already exists, the pipeline will not be created.
    This check can be disabled by adding the --allow-duplicate-names option.

    If the pipeline settings contain an "id" field, this command will fail.

    Usage:

    databricks pipelines create example.json

    OR

    databricks pipelines create --settings example.json
    """
    # pylint: enable=line-too-long
    if bool(settings_arg) == bool(settings):
        raise ValueError('Settings should be provided either as an argument ' +
                         '(databricks pipelines create example.json) or as ' +
                         'an option (databricks pipelines create --settings example.json).')

    src = settings_arg if bool(settings_arg) else settings
    settings_obj = _read_settings(src)
    settings_dir = os.path.dirname(src)

    if 'id' in settings_obj:
        raise ValueError("Pipeline settings shouldn't contain \"id\" "
                         "when creating a pipeline.")

    try:
        response = PipelinesApi(api_client).create(
            settings_obj, settings_dir, allow_duplicate_names)
    except requests.exceptions.HTTPError as e:
        _handle_duplicate_name_exception(settings_obj, e, is_create_pipeline=True)

    new_pipeline_id = response['pipeline_id']
    click.echo("Successfully created pipeline: {} with ID: {}.".format(
        _get_pipeline_url(api_client, new_pipeline_id), new_pipeline_id))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Edits a pipeline.')
@click.argument('settings_arg', default=None, required=False)
@click.option('--settings', default=None, type=PipelineSettingClickType(),
              help=PipelineSettingClickType.help)
@click.option('--pipeline-id', default=None, type=PipelineIdClickType(),
              help=PipelineIdClickType.help)
@click.option('--allow-duplicate-names', is_flag=True,
              help="Skip duplicate name check while editing pipeline.")
@debug_option
@profile_option
@pipelines_exception_eater
@provide_api_client
def edit_cli(api_client, settings_arg, settings, pipeline_id, allow_duplicate_names):
    # pylint: disable=line-too-long
    """
    Edits a pipeline specified by the pipeline settings. The pipeline settings are a
    JSON document that defines a Delta Live Tables pipeline on Databricks. To use a
    file containing the pipeline settings, pass the file path to the command as an
    argument or with the --settings option.

    Specification for the pipeline settings JSON can be found at
    https://docs.databricks.com/data-engineering/delta-live-tables/delta-live-tables-configuration.html

    If another pipeline with the same name exists, pipeline settings will not be edited.
    This check can be disabled by adding the --allow-duplicate-names option.

    Note that if an ID is specified in both the settings and passed with the --pipeline-id argument,
    the two ids must be the same, or the command will fail.

    Usage:

    databricks pipelines edit example.json

    OR

    databricks pipelines edit --settings example.json
    """
    # pylint: enable=line-too-long
    if bool(settings_arg) == bool(settings):
        raise ValueError('Settings should be provided either as an argument ' +
                         '(databricks pipelines edit example.json) or as ' +
                         'an option (databricks pipelines edit --settings example.json).')

    src = settings_arg if bool(settings_arg) else settings
    settings_obj = _read_settings(src)
    settings_dir = os.path.dirname(src)

    if (pipeline_id and 'id' in settings_obj) and pipeline_id != settings_obj["id"]:
        raise ValueError(
            "The ID provided in --pipeline_id '{}' is different from the ID provided "
            "in the settings '{}'. Resolve the conflict and try the command again. ".format(
                pipeline_id, settings_obj["id"])
        )

    settings_obj['id'] = pipeline_id or settings_obj.get('id', None)
    _validate_pipeline_id(settings_obj['id'])

    try:
        PipelinesApi(api_client).edit(settings_obj, settings_dir, allow_duplicate_names)
    except requests.exceptions.HTTPError as e:
        _handle_duplicate_name_exception(settings_obj, e, is_create_pipeline=False)
    click.echo("Successfully edited pipeline settings: {}.".format(
        _get_pipeline_url(api_client, settings_obj['id'])))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='[Deprecated] This command is deprecated, use create and edit '
                          'commands instead.\n Creates or edits a pipeline specified by the '
                          'pipeline settings.')
@click.argument('settings_arg', default=None, required=False)
@click.option('--settings', default=None, type=PipelineSettingClickType(),
              help=PipelineSettingClickType.help)
@click.option('--spec', default=None, type=PipelineSpecClickType(),
              help=PipelineSpecClickType.help)
@click.option('--allow-duplicate-names', is_flag=True,
              help="Skip duplicate name check while deploying pipeline.")
@click.option('--pipeline-id', default=None, type=PipelineIdClickType(),
              help=PipelineIdClickType.help)
@debug_option
@profile_option
@pipelines_exception_eater
@provide_api_client
def deploy_cli(api_client, settings_arg, settings, spec, allow_duplicate_names, pipeline_id):
    # pylint: disable=line-too-long
    """
    [Deprecated] This command is deprecated, use create and edit commands instead.

    Creates or edits a pipeline specified by the pipeline settings. The pipeline settings
    are a JSON document that defines a Delta Live Tables pipeline on Databricks. To use a
    file containing the pipeline settings, pass the file path to the command as an
    argument or with the --settings option.

    Specification for the pipeline settings JSON can be found at
    https://docs.databricks.com/data-engineering/delta-live-tables/delta-live-tables-configuration.html

    If the pipeline settings contains an "id" field, or if a pipeline ID is specified directly
    (using the  --pipeline-id argument), attempts to update an existing pipeline
    with that ID. If it does not, creates a new pipeline and logs the URL and the ID of the
    new pipeline to STDOUT. Note that if an ID is specified in both the settings and passed
    with the --pipeline-id argument, the two IDs must be the same, or the command will fail.

    The deploy command will not create a new pipeline if a pipeline with the same name already
    exists. This check can be disabled by adding the --allow-duplicate-names option.

    Usage:

    databricks pipelines deploy example.json

    OR

    databricks pipelines deploy --settings example.json

    OR

    databricks pipelines deploy --pipeline-id 1234 --settings example.json
    """
    # pylint: enable=line-too-long
    click.echo("DeprecationWarning: the \"deploy\" command is deprecated, " +
               "use \"create\" command to create a new pipeline or \"edit\" command " +
               "to modify an existing pipeline.\n")

    settings_error_msg = 'Settings should be provided either as an argument ' \
                         '(databricks pipelines deploy example.json) or as ' \
                         'an option (databricks pipelines deploy --settings example.json).'
    if bool(spec):
        if bool(spec) == bool(settings):
            raise ValueError(settings_error_msg)
        settings = spec

    if bool(settings_arg) == bool(settings):
        raise ValueError(settings_error_msg)

    src = settings_arg if bool(settings_arg) else settings
    settings_obj = _read_settings(src)
    settings_dir = os.path.dirname(src)
    if not pipeline_id and 'id' not in settings_obj:
        try:
            response = PipelinesApi(api_client).create(
                settings_obj, settings_dir, allow_duplicate_names)
        except requests.exceptions.HTTPError as e:
            _handle_duplicate_name_exception(settings_obj, e, is_create_pipeline=True)

        new_pipeline_id = response['pipeline_id']
        click.echo("Successfully created pipeline: {} with ID: {}".format(
            _get_pipeline_url(api_client, new_pipeline_id), new_pipeline_id))
    else:
        if (pipeline_id and 'id' in settings_obj) and pipeline_id != settings_obj["id"]:
            raise ValueError(
                "The ID provided in --pipeline_id '{}' is different from the ID provided "
                "in the settings '{}'. Resolve the conflict and try the command again.".format(
                    pipeline_id, settings_obj["id"])
            )

        settings_obj['id'] = pipeline_id or settings_obj.get('id', None)
        _validate_pipeline_id(settings_obj['id'])
        try:
            PipelinesApi(api_client).edit(
                settings_obj, settings_dir, allow_duplicate_names)
        except requests.exceptions.HTTPError as e:
            _handle_duplicate_name_exception(settings_obj, e, is_create_pipeline=False)
        click.echo("Successfully deployed pipeline: {}.".format(
            _get_pipeline_url(api_client, settings_obj['id'])))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Deletes the pipeline and cancels any active updates.')
@click.option('--pipeline-id', default=None, type=PipelineIdClickType(),
              help=PipelineIdClickType.help)
@debug_option
@profile_option
@pipelines_exception_eater
@provide_api_client
def delete_cli(api_client, pipeline_id):
    """
    Deletes the pipeline and cancels any active updates.

    Usage:

    databricks pipelines delete --pipeline-id 1234
    """
    _validate_pipeline_id(pipeline_id)
    PipelinesApi(api_client).delete(pipeline_id)
    click.echo("Pipeline {} deleted".format(pipeline_id))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Gets a pipeline\'s current settings and status.')
@click.option('--pipeline-id', default=None, type=PipelineIdClickType(),
              help=PipelineIdClickType.help)
@debug_option
@profile_option
@pipelines_exception_eater
@provide_api_client
def get_cli(api_client, pipeline_id):
    """
    Gets a pipeline's current settings and status.

    Usage:

    databricks pipelines get --pipeline-id 1234
    """
    _validate_pipeline_id(pipeline_id)
    click.echo(pretty_format(PipelinesApi(api_client).get(pipeline_id)))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Lists all pipelines and their statuses.')
@debug_option
@profile_option
@pipelines_exception_eater
@provide_api_client
def list_cli(api_client):
    """
    Lists all pipelines and their statuses.

    Usage:

    databricks pipelines list
    """
    click.echo(pretty_format(PipelinesApi(api_client).list()))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='[Deprecated] Use the "start --full-refresh" command instead. ' +
                          'Resets a pipeline so that data can be reprocessed ' +
                          'from the beginning.')
@click.option('--pipeline-id', default=None, type=PipelineIdClickType(),
              help=PipelineIdClickType.help)
@debug_option
@profile_option
@pipelines_exception_eater
@provide_api_client
def reset_cli(api_client, pipeline_id):
    """
    [Deprecated] Use the "start --full-refresh" command instead.

    Resets a pipeline by truncating tables and creating new checkpoint folders so that data is
    reprocessed from the beginning.

    Usage:

    databricks pipelines reset --pipeline-id 1234
    """
    click.echo("DeprecationWarning: the \"reset\" command is deprecated, " +
               "use the \"start --full-refresh\" command instead.")
    _validate_pipeline_id(pipeline_id)
    resp = PipelinesApi(api_client).start_update(pipeline_id, full_refresh=True)
    click.echo(_gen_start_update_msg(resp, pipeline_id, True))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='[Deprecated] Use the "start" command instead. ' +
                          'Starts a pipeline update.')
@click.option('--pipeline-id', default=None, type=PipelineIdClickType(),
              help=PipelineIdClickType.help)
@debug_option
@profile_option
@pipelines_exception_eater
@provide_api_client
def run_cli(api_client, pipeline_id):
    """
    [Deprecated] Use the "start" command instead.

    Starts a pipeline update.

    Usage:

    databricks pipelines run --pipeline-id 1234
    """
    click.echo("Deprecation warning: the \"run\" command is deprecated." +
               " Use the \"start\" command instead.")
    _validate_pipeline_id(pipeline_id)
    resp = PipelinesApi(api_client).start_update(pipeline_id, full_refresh=False)
    click.echo(_gen_start_update_msg(resp, pipeline_id, False))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Starts a pipeline update.')
@click.option('--pipeline-id', default=None, type=PipelineIdClickType(),
              help=PipelineIdClickType.help)
@click.option('--full-refresh', default=False, type=bool, is_flag=True,
              help='If present, truncates tables and creates new checkpoint ' +
                   'folders so that data is reprocessed from the beginning.')
@debug_option
@profile_option
@pipelines_exception_eater
@provide_api_client
def start_cli(api_client, pipeline_id, full_refresh):
    """
    Starts a pipeline update.

    Usage:

    databricks pipelines start --pipeline-id 1234 --full-refresh
    """
    _validate_pipeline_id(pipeline_id)
    resp = PipelinesApi(api_client).start_update(pipeline_id, full_refresh=full_refresh)
    click.echo(_gen_start_update_msg(resp, pipeline_id, full_refresh))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Stops the pipeline by cancelling any active update.')
@click.option('--pipeline-id', default=None, type=PipelineIdClickType(),
              help=PipelineIdClickType.help)
@debug_option
@profile_option
@pipelines_exception_eater
@provide_api_client
def stop_cli(api_client, pipeline_id):
    """
    Stops the pipeline by cancelling any active update.

    Usage:

    databricks pipelines stop --pipeline-id 1234
    """
    _validate_pipeline_id(pipeline_id)
    PipelinesApi(api_client).stop(pipeline_id)
    click.echo("Stopped pipeline {}.".format(pipeline_id))


def _gen_start_update_msg(resp, pipeline_id, full_refresh):
    output_msg = "Started an update "
    if resp and 'update_id' in resp:
        output_msg += "{} ".format(resp.get('update_id'))

    if full_refresh:
        output_msg += "with full refresh "

    output_msg += "for pipeline {}.".format(pipeline_id)
    return output_msg


def _read_settings(src):
    """
    Reads the settings at src as a JSON If the file has JSON extension or
    if no file extension is provided. Other file extensions and formats are
    not supported.
    """
    extension = os.path.splitext(src)[1]
    if extension.lower() == '.json' or (not extension):
        try:
            with open(src, 'r') as f:
                data = f.read()
            return json.loads(data)
        except json_parse_exception as e:
            error_and_quit("Invalid JSON provided in settings\n{}.".format(e))
    else:
        raise ValueError('The provided file extension for the settings is not supported. ' +
                         'Only JSON files are supported.')


def _get_pipeline_url(api_client, pipeline_id):
    base_url = "{0.scheme}://{0.netloc}/".format(urlparse(api_client.url))
    return urljoin(base_url, "#joblist/pipelines/{}".format(pipeline_id))


def _validate_pipeline_id(pipeline_id):
    """
    Checks if the pipeline ID is not empty.
    """
    if pipeline_id is None or len(pipeline_id) == 0:
        error_and_quit(u'Empty pipeline ID provided')


def _handle_duplicate_name_exception(settings, exception, is_create_pipeline):
    error_code = None
    try:
        error_code = json.loads(exception.response.text).get('error_code')
    except ValueError:
        pass

    if error_code == 'RESOURCE_CONFLICT':
        if is_create_pipeline:
            raise ValueError(
                "Pipeline with name '{}' already exists. ".format(settings['name']) +
                "If you are updating an existing pipeline, use \"edit\" command. "
                "Otherwise, You can use the --allow-duplicate-names option to skip "
                "this check. ")

        raise ValueError(
            "Pipeline with name '{}' already exists. ".format(settings['name']) +
            "You can use the --allow-duplicate-names option to skip this check. ")

    raise exception


@click.group(context_settings=CONTEXT_SETTINGS,
             short_help='Utility to interact with Databricks Delta Live Tables Pipelines.')
@click.option('--version', '-v', is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=version)
@debug_option
@profile_option
def pipelines_group():  # pragma: no cover
    """
    Utility to interact with Databricks Delta Live Tables Pipelines.
    """
    pass


pipelines_group.add_command(deploy_cli, name='deploy')
pipelines_group.add_command(create_cli, name="create")
pipelines_group.add_command(edit_cli, name="edit")
pipelines_group.add_command(delete_cli, name='delete')
pipelines_group.add_command(get_cli, name='get')
pipelines_group.add_command(list_cli, name='list')
pipelines_group.add_command(start_cli, name='start')
pipelines_group.add_command(stop_cli, name='stop')

# DEPRECATED and will be removed in future versions.
pipelines_group.add_command(reset_cli, name='reset')
pipelines_group.add_command(run_cli, name='run')



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import requests
from databricks_cli.sdk import ReposService, WorkspaceService


class ReposApi(object):
    def __init__(self, api_client):
        self.client = ReposService(api_client)
        self.ws_client = WorkspaceService(api_client)

    def get_repo_id(self, path):
        if not path.startswith("/Repos/"):
            raise ValueError("Path must start with /Repos/ !")

        if not len([x for x in path.split("/") if x]) == 3:
            raise ValueError("Repos paths must be in /Repos/{folder}/{repo} format!")

        try:
            status = self.ws_client.get_status(path)
            if status['object_type'] == 'REPO':
                return status['object_id']
        except requests.exceptions.HTTPError as ex:
            if ex.response.status_code != 404:
                jsn = ex.response.json()
                if 'message' in jsn:
                    msg = jsn['message']
                else:
                    msg = ex.response.reason
                raise RuntimeError(
                    "Error fetching repo ID for {path}: HTTP code: {code}, {ex}".format(
                        path=path, code=ex.response.status_code, ex=msg))

        raise RuntimeError("Can't find repo ID for {path}".format(path=path))

    def list(self, path_prefix, next_page_token):
        """
        List repos that the caller has Manage permissions on. Results are
        paginated with each page containing twenty repos.
        """
        return self.client.list_repos(path_prefix, next_page_token)

    def create(self, url, provider, path):
        """
        Creates a repo object and links it to the remote Git repo specified.
        """
        return self.client.create_repo(url, provider, path)

    def get(self, repo_id):
        """
        Gets the repo with the given ID.
        """
        return self.client.get_repo(repo_id)

    def update(self, repo_id, branch, tag):
        """
        Checks out the repo to the given branch or tag. Only one of ``branch``
        or ``tag`` should be provided.
        """
        assert bool(branch is not None) ^ bool(tag is not None)
        return self.client.update_repo(repo_id, branch, tag)

    def delete(self, repo_id):
        """
        Deletes the repo with the given ID.
        """
        return self.client.delete_repo(repo_id)



#!/usr/bin/env python

# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
A common class to be used by client of different APIs
"""

import base64
import json
import warnings
import requests
import ssl
import copy
import pprint

from . import version

from requests.adapters import HTTPAdapter
from requests.utils import get_netrc_auth
from requests.auth import HTTPBasicAuth
from six.moves.urllib.parse import urlparse

try:
    from requests.packages.urllib3.poolmanager import PoolManager
    from requests.packages.urllib3 import exceptions
    from requests.packages.urllib3.util.retry import Retry
except ImportError:
    from urllib3.poolmanager import PoolManager
    from urllib3 import exceptions
    from urllib3.util.retry import Retry

from databricks_cli.sdk.version import UC_API_VERSION
from databricks_cli.version import version as databricks_cli_version

class TlsV1HttpAdapter(HTTPAdapter):
    """
    A HTTP adapter implementation that specifies the ssl version to be TLS1.
    This avoids problems with openssl versions that
    use SSL3 as a default (which is not supported by the server side).
    """

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = PoolManager(num_pools=connections, maxsize=maxsize, block=block, ssl_version=ssl.PROTOCOL_TLSv1_2)

# https://github.com/psf/requests/issues/2773#issuecomment-174312831
class FallbackNetrcAuth(requests.auth.AuthBase):
    '''Force requests to ignore the ``.netrc`` if other authentication 
    methods have been setup. Fallback to ``.netrc`` if not. 

    Use with::

        requests.get(url, auth=FallbackNetrcAuth())
        
        s = requests.Session()
        s.auth = FallbackNetrcAuth()
    '''

    def __call__(self, r):
        if "Authorization" in r.headers:
            return r
        
        netrc_tuple = get_netrc_auth(r.url)

        if netrc_tuple is None or not any(netrc_tuple):
            return r
        
        return HTTPBasicAuth(*netrc_tuple)(r)

class ApiClient(object):
    """
    A partial Python implementation of dbc rest api
    to be used by different versions of the client.
    """
    def __init__(self, user=None, password=None, host=None, token=None,
                 api_version=version.API_VERSION, default_headers={}, verify=True, command_name="", jobs_api_version=None):
        if host[-1] == "/":
            host = host[:-1]

        retries = Retry(
            total=6,
            backoff_factor=1,
            status_forcelist=[429],
            method_whitelist=set({'POST'}) | set(Retry.DEFAULT_METHOD_WHITELIST),
            respect_retry_after_header=True,
            raise_on_status=False # return original response when retries have been exhausted
        )
        self.session = requests.Session()
        self.session.auth = FallbackNetrcAuth()
        self.session.mount('https://', TlsV1HttpAdapter(max_retries=retries))

        parsed_url = urlparse(host)
        scheme = parsed_url.scheme
        hostname = parsed_url.hostname
        self.url = "%s://%s/api/" % (scheme, hostname)
        if user is not None and password is not None:
            encoded_auth = (user + ":" + password).encode()
            user_header_data = "Basic " + base64.standard_b64encode(encoded_auth).decode()
            auth = {'Authorization': user_header_data, 'Content-Type': 'text/json'}
        elif token is not None:
            auth = {'Authorization': 'Bearer {}'.format(token), 'Content-Type': 'text/json'}
        else:
            auth = {}
        user_agent = {'user-agent': 'databricks-cli-{v}-{c}'.format(v=databricks_cli_version,
                                                                    c=command_name)}
        self.default_headers = {}
        self.default_headers.update(auth)
        self.default_headers.update(default_headers)
        self.default_headers.update(user_agent)
        self.verify = verify
        self.api_version = api_version
        self.jobs_api_version = jobs_api_version

    def close(self):
        """Close the client"""
        pass

    # helper functions starting here

    def perform_query(self, method, path, data = {}, headers = None, files=None, version=None):
        """set up connection and perform query"""
        if headers is None:
            headers = self.default_headers
        else:
            tmp_headers = copy.deepcopy(self.default_headers)
            tmp_headers.update(headers)
            headers = tmp_headers

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", exceptions.InsecureRequestWarning)
            if method == 'GET':
                translated_data = {k: _translate_boolean_to_query_param(data[k]) for k in data}
                resp = self.session.request(method, self.get_url(path, version=version), params = translated_data,
                                            verify = self.verify, headers = headers)
            else:
                if files is None:
                    resp = self.session.request(method, self.get_url(path, version=version), data = json.dumps(data),
                                                verify = self.verify, headers = headers)
                else:
                    # Multipart file upload
                    resp = self.session.request(method, self.get_url(path, version=version), files = files, data = data,
                                                verify = self.verify, headers = headers)
        try:
            resp.raise_for_status()
        except requests.exceptions.HTTPError as e:
            message = e.args[0]
            try:
                reason = pprint.pformat(json.loads(resp.text), indent=2)
                message += '\n Response from server: \n {}'.format(reason)
            except ValueError:
                pass
            raise requests.exceptions.HTTPError(message, response=e.response)
        return resp.json()


    def get_url(self, path, version=None):
        if version:
            return self.url + version + path
        elif self.jobs_api_version and path and path.startswith('/jobs'):
            return self.url + self.jobs_api_version + path
        elif path and _is_uc_path(path):
            return self.url + UC_API_VERSION + path
        return self.url + self.api_version + path


def _is_uc_path(path):
    return path.startswith('/unity-catalog')


def _translate_boolean_to_query_param(value):
    assert not isinstance(value, list), 'GET parameters cannot pass list of objects'
    if isinstance(value, bool):
        if value:
            return 'true'
        else:
            return 'false'
    return value



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
from tabulate import tabulate

from databricks_cli.click_types import OutputClickType, JsonClickType, InstancePoolIdClickType
from databricks_cli.instance_pools.api import InstancePoolsApi
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, pretty_format, json_cli_base, \
    truncate_string
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.version import print_version_callback, version


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--json-file', default=None, type=click.Path(),
              help='File containing JSON request to POST to /api/2.0/instance-pools/create.')
@click.option('--json', default=None, type=JsonClickType(),
              help=JsonClickType.help('/api/2.0/instance-pools/create'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def create_cli(api_client, json_file, json):
    """
    Creates a Databricks instance pool.

    The specification for the request json can be found at
    https://docs.databricks.com/api/latest/instance-pools.html#create
    """
    json_cli_base(json_file, json,
                  lambda json: InstancePoolsApi(api_client).create_instance_pool(json))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--json-file', default=None, type=click.Path(),
              help='File containing JSON request to POST to /api/2.0/instance-pools/edit.')
@click.option('--json', default=None, type=JsonClickType(),
              help=JsonClickType.help('/api/2.0/instance-pools/edit'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def edit_cli(api_client, json_file, json):
    """
    Edits a Databricks instance pool.

    The specification for the request json can be found at
    https://docs.databricks.com/api/latest/instance-pools.html#edit
    """
    if not bool(json_file) ^ bool(json):
        raise RuntimeError('Either --json-file or --json should be provided')
    json_cli_base(json_file, json,
                  lambda json: InstancePoolsApi(api_client).edit_instance_pool(json),
                  print_response=False)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--instance-pool-id', required=True, type=InstancePoolIdClickType(),
              help=InstancePoolIdClickType.help)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_cli(api_client, instance_pool_id):
    """
    Deletes a Databricks instance pool given its ID.

    This permanently deletes the instance pool. The idle instances in the pool are terminated
    asynchronously. New clusters cannot attach to the pool. Running clusters attached to the pool
    continue to run but cannot auto-scale up. Terminated clusters attached to the pool will fail to
    start until they are edited to no longer use the pool.
    """
    InstancePoolsApi(api_client).delete_instance_pool(instance_pool_id)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--instance-pool-id', required=True, type=InstancePoolIdClickType(),
              help=InstancePoolIdClickType.help)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_cli(api_client, instance_pool_id):
    """
    Retrieves metadata about an instance pool.
    """
    click.echo(pretty_format(InstancePoolsApi(api_client).get_instance_pool(instance_pool_id)))


def _instance_pools_to_table(instance_pools_json):
    ret = []
    stats_headers = ['idle_count', 'used_count', 'pending_idle_count', 'pending_used_count']
    for c in instance_pools_json.get('instance_pools', []):
        pool_stats = []
        pool_stats.append(c['instance_pool_id'])
        pool_stats.append(truncate_string(c['instance_pool_name']))
        for header in stats_headers:
            pool_stats.append(c['stats'][header])
        # clone the content in the pool_stats. Pool_stats will be re-used in next iteration.
        ret.append(pool_stats[:])
    return ret


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Lists active and recently terminated instance pools.')
@click.option('--output', default=None, help=OutputClickType.help, type=OutputClickType())
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_cli(api_client, output):
    """
    Lists active instance pools with the stats of the pools.
    """
    instance_pools_json = InstancePoolsApi(api_client).list_instance_pools()
    if OutputClickType.is_json(output):
        click.echo(pretty_format(instance_pools_json))
    else:
        headers = ['ID', 'NAME', 'IDLE INSTANCES', 'USED INSTANCES', 'PENDING IDLE INSTANCES',
                   'PENDING USED INSTANCES']
        click.echo(tabulate(_instance_pools_to_table(instance_pools_json), headers=headers,
                            tablefmt='plain', numalign='left'))


@click.group(context_settings=CONTEXT_SETTINGS,
             short_help='Utility to interact with Databricks instance pools.')
@click.option('--version', '-v', is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=version)
@debug_option
@profile_option
@eat_exceptions
def instance_pools_group():  # pragma: no cover
    """
    Utility to interact with Databricks instance pools.
    """
    pass


instance_pools_group.add_command(create_cli, name='create')
instance_pools_group.add_command(edit_cli, name='edit')
instance_pools_group.add_command(delete_cli, name='delete')
instance_pools_group.add_command(get_cli, name='get')
instance_pools_group.add_command(list_cli, name='list')



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import time
from json import loads as json_loads

import click
from tabulate import tabulate

from databricks_cli.click_types import OutputClickType, JsonClickType, RunIdClickType
from databricks_cli.jobs.cli import check_version
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, pretty_format, truncate_string, \
    error_and_quit, backoff_with_jitter
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option, \
    api_version_option
from databricks_cli.runs.api import RunsApi
from databricks_cli.version import print_version_callback, version as cli_version


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--json-file', default=None, type=click.Path(),
              help='File containing JSON request to POST to /api/2.*/jobs/runs/submit.')
@click.option('--json', default=None, type=JsonClickType(),
              help=JsonClickType.help('/api/2.*/jobs/runs/submit'))
@click.option('--wait', is_flag=True, default=False,
              help='Waits for the submitted run to complete.')
@api_version_option
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def submit_cli(api_client, json_file, json, wait, version):
    """
    Submits a one-time run and optionally waits for its completion.

    The specification for the request json can be found
    https://docs.databricks.com/api/latest/jobs.html#runs-submit
    """
    check_version(api_client, version)
    if json_file:
        with open(json_file, 'r') as f:
            json = f.read()
    submit_res = RunsApi(api_client).submit_run(json_loads(json), version=version)
    click.echo(pretty_format(submit_res))
    if wait:
        run_id = submit_res['run_id']
        completed_states = set(['TERMINATED', 'SKIPPED', 'INTERNAL_ERROR'])
        prev_life_cycle_state = ""
        attempt = 0
        # Wait for run to complete
        while True:
            run = RunsApi(api_client).get_run(run_id, version=version)
            run_state = run['state']
            life_cycle_state = run_state['life_cycle_state']
            if life_cycle_state in completed_states:
                if run_state['result_state'] == 'SUCCESS':
                    sys.exit(0)
                else:
                    error_and_quit('Run failed with state ' + run_state['result_state'] +
                                   ' and state message ' + run_state['state_message'])
            if prev_life_cycle_state != life_cycle_state:
                click.echo('Waiting on run to complete. Current state: ' + life_cycle_state +
                           '. URL: ' + run['run_page_url'], err=True)
                prev_life_cycle_state = life_cycle_state
            time.sleep(backoff_with_jitter(attempt))
            attempt += 1


def _runs_to_table(runs_json):
    ret = []
    for r in runs_json.get('runs', []):
        run_id = r.get('run_id', 'no_run_id')
        run_name = r.get('run_name', 'no_run_name')
        life_cycle_state = r.get('state', {}).get('life_cycle_state', 'n/a')
        result_state = r.get('state', {}).get('result_state', 'n/a')
        run_page_url = r.get('run_page_url', 'n/a')
        row = (run_id, truncate_string(run_name),
               life_cycle_state, result_state, run_page_url)
        ret.append(row)
    return ret


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--job-id', default=None, type=int,
              help='If specified, runs from only the specified job_id will be listed.')
@click.option('--active-only', is_flag=True, default=None,
              help='If specified, only active runs will be listed')
@click.option('--completed-only', is_flag=True, default=None,
              help='If specified, only completed runs will be listed')
@click.option('--offset', default=None, type=int,
              help='The offset is relative to the most recent run ID. Set to 0 by default.')
@click.option('--limit', default=None, type=int,
              help='The limit determines the number of runs listed. '
                   'Limit must be between 0 and 1000. Set to 20 runs by default.')
@click.option('--output', help=OutputClickType.help, type=OutputClickType())
@api_version_option
@debug_option
@profile_option
@eat_exceptions  # noqa
@provide_api_client
def list_cli(api_client, job_id, active_only, completed_only, offset, limit, output, version):  # noqa
    """
    Lists job runs.

    The limit and offset determine which runs will be listed. Runs are always listed
    by descending order of run start time and run ID.

    In the TABLE output mode, the columns are as follows.

      - Run ID

      - Run name

      - Life cycle state

      - Result state (can be n/a)
    """
    check_version(api_client, version)
    runs_json = RunsApi(api_client).list_runs(
        job_id, active_only, completed_only, offset, limit, version=version)
    if OutputClickType.is_json(output):
        click.echo(pretty_format(runs_json))
    else:
        click.echo(tabulate(_runs_to_table(runs_json), tablefmt='plain'))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--run-id', required=True, type=RunIdClickType())
@api_version_option
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_cli(api_client, run_id, version):
    """
    Gets the metadata about a run in json form.

    The output schema is documented https://docs.databricks.com/api/latest/jobs.html#runs-get.
    """
    check_version(api_client, version)
    click.echo(pretty_format(
        RunsApi(api_client).get_run(run_id, version=version)))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--run-id', required=True, type=RunIdClickType())
@api_version_option
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_output_cli(api_client, run_id, version):
    """
    Gets the output of a run

    The output schema is documented https://docs.databricks.com/api/latest/jobs.html#runs-get-output
    """
    check_version(api_client, version)
    click.echo(pretty_format(
        RunsApi(api_client).get_run_output(run_id, version=version)))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--run-id', required=True, type=RunIdClickType())
@api_version_option
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def cancel_cli(api_client, run_id, version):
    """
    Cancels the run specified.
    """
    check_version(api_client, version)
    click.echo(pretty_format(
        RunsApi(api_client).cancel_run(run_id, version=version)))


@click.group(context_settings=CONTEXT_SETTINGS,
             short_help='Utility to interact with the jobs runs.')
@click.option('--version', '-v', is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=cli_version)
@debug_option
@profile_option
@eat_exceptions
def runs_group():  # pragma: no cover
    """
    Utility to interact with jobs runs.
    """
    pass


runs_group.add_command(submit_cli, name='submit')
runs_group.add_command(list_cli, name='list')
runs_group.add_command(get_cli, name='get')
runs_group.add_command(cancel_cli, name='cancel')
runs_group.add_command(get_output_cli, name='get-output')



# Databricks CLI
# Copyright 2022 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from databricks_cli.click_types import JsonClickType
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.unity_catalog.api import UnityCatalogApi
from databricks_cli.unity_catalog.utils import hide, json_file_help, json_string_help, \
    mc_pretty_format
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, json_cli_base


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Create a table. [DO NOT USE]',
               hidden=True)
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='POST', path='/tables'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='POST', path='/tables'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def create_table_cli(api_client, json_file, json):
    """
    Create new table specified by the JSON input.

    WARNING: Creating table metadata via the UC API may create a table
    that is unusable in DBR. Instead, use SQL commands (CREATE TABLE) in DBR.

    The public specification for the JSON request is in development.
    """
    json_cli_base(json_file, json,
                  lambda json: UnityCatalogApi(api_client).create_table(json),
                  encode_utf8=True)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List tables.')
@click.option('--catalog-name', required=True,
              help='Name of the parent catalog for tables of interest.')
@click.option('--schema-name', required=True,
              help='Name of the parent schema for tables of interest.')
@click.option('--name-pattern', default=None,
              help='SQL LIKE pattern that the table name must match to be in list.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_tables_cli(api_client, catalog_name, schema_name, name_pattern):
    """
    List tables.
    """
    tables_json = UnityCatalogApi(api_client).list_tables(catalog_name, schema_name, name_pattern)
    click.echo(mc_pretty_format(tables_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List table summaries.')
@click.option('--catalog-name', required=True,
              help='Name of the parent catalog for tables of interest.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_table_summaries_cli(api_client, catalog_name):
    """
    List table summaries (in bulk).
    """
    tables_json = UnityCatalogApi(api_client).list_table_summaries(catalog_name)
    click.echo(mc_pretty_format(tables_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Get a table.')
@click.option('--full-name', required=True,
              help='Full name (<catalog>.<schema>.<table>) of the table to get.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_table_cli(api_client, full_name):
    """
    Get a table.
    """
    table_json = UnityCatalogApi(api_client).get_table(full_name)
    click.echo(mc_pretty_format(table_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Update a table. [DO NOT USE]',
               hidden=True)
@click.option('--full-name', required=True,
              help='Full name (<catalog>.<schema>.<table>) of the table to update.')
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='PATCH', path='/tables/{name}'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='PATCH', path='/tables/{name}'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def update_table_cli(api_client, full_name, json_file, json):
    """
    Update a table.

    WARNING: Altering table metadata via the UC API may cause the table
    to be unusable in DBR. Instead, use SQL commands (ALTER TABLE) in DBR.

    The public specification for the JSON request is in development.
    """
    json_cli_base(json_file, json,
                  lambda json: UnityCatalogApi(api_client).update_table(full_name, json),
                  encode_utf8=True)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Delete a table.')
@click.option('--full-name', required=True,
              help='Full name (<catalog>.<schema>.<table>) of the table to delete.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_table_cli(api_client, full_name):
    """
    Delete a table.
    """
    UnityCatalogApi(api_client).delete_table(full_name)


@click.group(context_settings=CONTEXT_SETTINGS,
             short_help=' ',
             help='Note:\n' +
                  'To create or update tables, use SQL commands\n' +
                  '(CREATE TABLE or ALTER TABLE) on a cluster or SQL warehouse.')
def tables_group():  # pragma: no cover
    pass


def register_table_commands(cmd_group):
    # Register deprecated "verb-noun" commands for backward compatibility.
    cmd_group.add_command(hide(create_table_cli), name='create-table')
    cmd_group.add_command(hide(list_tables_cli), name='list-tables')
    cmd_group.add_command(hide(list_table_summaries_cli), name='list-table-summaries')
    cmd_group.add_command(hide(get_table_cli), name='get-table')
    cmd_group.add_command(hide(update_table_cli), name='update-table')
    cmd_group.add_command(hide(delete_table_cli), name='delete-table')

    # Register command group.
    tables_group.add_command(create_table_cli, name='create')
    tables_group.add_command(list_tables_cli, name='list')
    tables_group.add_command(list_table_summaries_cli, name='list-summaries')
    tables_group.add_command(get_table_cli, name='get')
    tables_group.add_command(update_table_cli, name='update')
    tables_group.add_command(delete_table_cli, name='delete')
    cmd_group.add_command(tables_group, name='tables')



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re


version = '0.17.6' #  NOQA


def print_version_callback(ctx, param, value): #  NOQA
    import click
    if not value or ctx.resilient_parsing:
        return
    click.echo('Version {}'.format(version))
    ctx.exit()


def _match_version(value):
    # Expect version to be of the form: `X.Y.Z(.suffix)?`
    match = re.match(r'^(\d+)\.(\d+)\.(\d+)(\.\w+)?$', value)
    if match is None:
        raise ValueError("Non-compliant version string: " + value)
    return match


def is_release_version(value=None):
    """
    Returns whether the current version of databricks-cli is a release version or not.
    """
    if value is None:
        value = version

    # The 4th group is the optional `.devZZZ` suffix.
    # If it is non-empty, this is not a release version.
    match = _match_version(value)
    if match.group(4) is not None:
        return False

    return True


def next_development_version(value=None):
    """
    Returns the hypothetical next development version of databricks-cli.
    """
    if value is None:
        value = version

    match = _match_version(value)
    major = int(match.group(1))
    minor = int(match.group(2))
    patch = int(match.group(3))
    return "{}.{}.{}.dev0".format(major, minor, patch + 1)



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import random
import sys
import traceback
from json import dumps as json_dumps, loads as json_loads

import click
import six
from requests.exceptions import HTTPError

from databricks_cli.click_types import ContextObject

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
CLUSTER_OPTIONS = ['cluster-id', 'cluster-name']
DEBUG_MODE = False


def eat_exceptions(function):
    @six.wraps(function)
    def decorator(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except HTTPError as exception:
            if exception.response.status_code == 401:
                error_and_quit('Your authentication information may be incorrect. Please '
                               'reconfigure with ``dbfs configure``')
            elif exception.response.status_code == 403:
                error_and_quit('Authorization failed. Your token may be expired or '
                               'lack the valid scope')
            else:
                error_and_quit(exception.response.content)
        except Exception as exception:  # noqa
            if not DEBUG_MODE:
                error_and_quit('{}: {}'.format(type(exception).__name__, str(exception)))

    decorator.__doc__ = function.__doc__
    return decorator


def pipelines_exception_eater(function):
    """
    Formats error messages from the pipelines API while keeping the existing
    behavior of eat_exception
    """

    @six.wraps(function)
    def decorator(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except HTTPError as exception:  # noqa
            if exception.response.status_code == 401:
                error_and_quit('Your authentication information may be incorrect. Please '
                               + 'reconfigure with ``dbfs configure``')
            else:
                try:
                    exp_context = json_loads(exception.response.content.decode('utf-8'))
                    message = exception.response.content
                    if 'error_code' in exp_context and 'message' in exp_context:
                        message = exp_context['error_code'] + '\n' + exp_context['message']
                    error_and_quit(message)
                except Exception:  # noqa
                    error_and_quit(exception.response.content)
        except Exception as exception:  # noqa
            if not DEBUG_MODE:
                error_and_quit('{}: {}'.format(type(exception).__name__, str(exception)))

    decorator.__doc__ = function.__doc__
    return decorator


def error_and_quit(message):
    ctx = click.get_current_context()
    context_object = ctx.ensure_object(ContextObject)
    if context_object.debug_mode:
        traceback.print_exc()
    click.echo(u'Error: {}'.format(message))
    sys.exit(1)


INTERVAL_MAX = 30
INTERVAL_BASE = 5
MAX_EXPONENT = 10


def backoff_with_jitter(attempt):
    """
    Creates a growing but randomized wait time based on the number of attempts already made.
    """
    exponent = min(attempt, MAX_EXPONENT)
    sleep_time = min(INTERVAL_MAX, INTERVAL_BASE * 2 ** exponent)
    return random.randrange(math.floor(sleep_time * 0.5), sleep_time)


def pretty_format(json, encode_utf8=False):
    if encode_utf8:
        return json_dumps(json, indent=2, ensure_ascii=False)
    return json_dumps(json, indent=2)


def json_cli_base(json_file, json, api, error_msg='', print_response=True, encode_utf8=False):
    """
    Takes json_file or json string and calls an function "api" with the json
    deserialized
    """
    if not (json_file is None) ^ (json is None):
        if not error_msg:
            error_msg = 'Either --json-file or --json should be provided'
        raise RuntimeError(error_msg)
    if json_file:
        with open(json_file, 'r') as f:
            json = f.read()
    res = api(json_loads(json))
    if print_response:
        click.echo(pretty_format(res, encode_utf8))


def truncate_string(s, length=100):
    if len(s) <= length:
        return s
    return s[:length] + '...'


def to_graph(dict_obj, graph_name="Graph"):
    entries = ["\"{}\" -> {};".format(key, ','.join(
        '"{}"'.format(item) for item in dict_obj.get(key))) for key in dict_obj]
    return "digraph \"{}\" ".format(graph_name) + "{\n\t" + "\n\t".join(entries) + "\n}"


class InvalidConfigurationError(RuntimeError):
    @staticmethod
    def for_profile(profile):
        if profile is None:
            return InvalidConfigurationError(
                'You haven\'t configured the CLI yet! '
                'Please configure by entering `{} configure`'.format(sys.argv[0]))
        return InvalidConfigurationError(
            ('You haven\'t configured the CLI yet for the profile {profile}! '
             'Please configure by entering '
             '`{argv} configure --profile {profile}`').format(
                profile=profile, argv=sys.argv[0]))


def merge_dicts_shallow(*dicts):
    """
    Merges dicts through shallow copy.
    """
    result = {}
    for d in dicts:
        result.update(d)
    return result
    



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
from tabulate import tabulate

from databricks_cli.utils import eat_exceptions, error_and_quit, CONTEXT_SETTINGS
from databricks_cli.version import print_version_callback, version
from databricks_cli.configure.cli import configure_cli
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.dbfs.api import DbfsApi
from databricks_cli.dbfs.dbfs_path import DbfsPath, DbfsPathClickType


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--absolute', is_flag=True, default=False,
              help='Displays absolute paths.')
@click.option('-l', is_flag=True, default=False,
              help="""Displays full information including size, file type 
                      and modification time since Epoch in milliseconds.""")
@click.argument('dbfs_path', nargs=-1, type=DbfsPathClickType())
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def ls_cli(api_client, l, absolute, dbfs_path): #  NOQA
    """
    List files in DBFS.
    """
    if len(dbfs_path) == 0:
        dbfs_path = DbfsPath('dbfs:/')
    elif len(dbfs_path) == 1:
        dbfs_path = dbfs_path[0]
    else:
        error_and_quit('ls can take a maximum of one path.')
    files = DbfsApi(api_client).list_files(dbfs_path)
    table = tabulate([f.to_row(is_long_form=l, is_absolute=absolute) for f in files],
                     tablefmt='plain')
    click.echo(table)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('dbfs_path', type=DbfsPathClickType())
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def mkdirs_cli(api_client, dbfs_path):
    """
    Make directories in DBFS.

    Mkdirs will create directories along the path to the argument directory.
    """
    DbfsApi(api_client).mkdirs(dbfs_path)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--recursive', '-r', is_flag=True, default=False)
@click.argument('dbfs_path', type=DbfsPathClickType())
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def rm_cli(api_client, recursive, dbfs_path):
    """
    Remove files from dbfs.

    To remove a directory you must provide the --recursive flag.
    """
    DbfsApi(api_client).delete(dbfs_path, recursive)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--recursive', '-r', is_flag=True, default=False)
@click.option('--overwrite', is_flag=True, default=False)
@click.argument('src')
@click.argument('dst')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def cp_cli(api_client, recursive, overwrite, src, dst):
    """
    Copy files to and from DBFS.

    Note that this function will fail if the src and dst are both on the local filesystem.

    For non-recursive copies, if the dst is a directory, the file will be placed inside the
    directory. For example ``dbfs cp dbfs:/apple.txt .`` will create a file at `./apple.txt`.

    For recursive copies, files inside of the src directory will be copied inside the dst directory
    with the same name. If the dst path does not exist, a directory will be created. For example
    ``dbfs cp -r dbfs:/foo foo`` will create a directory foo and place the files ``dbfs:/foo/a`` at
    ``foo/a``. If ``foo/a`` already exists, the file will not be overridden unless the --overwrite
    flag is provided -- however, dbfs cp --recursive will continue to try and copy other files.
    """
    # Copy to DBFS in this case
    DbfsApi(api_client).cp(recursive, overwrite, src, dst)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('src', type=DbfsPathClickType())
@click.argument('dst', type=DbfsPathClickType())
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def mv_cli(api_client, src, dst):
    """
    Moves a file between two DBFS paths.
    """
    DbfsApi(api_client).move(src, dst)


@click.group(context_settings=CONTEXT_SETTINGS, short_help='Utility to interact with DBFS.')
@click.option('--version', '-v', is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=version)
@debug_option
@profile_option
def dbfs_group():  # pragma: no cover
    """
    Utility to interact with DBFS.

    DBFS paths are all prefixed with dbfs:/. Local paths can be absolute or local.
    """
    pass


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('src')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def cat_cli(api_client, src):
    """
    Show the contents of a file. Does not work for directories.
    """
    DbfsApi(api_client).cat(src)


dbfs_group.add_command(configure_cli, name='configure')
dbfs_group.add_command(ls_cli, name='ls')
dbfs_group.add_command(mkdirs_cli, name='mkdirs')
dbfs_group.add_command(rm_cli, name='rm')
dbfs_group.add_command(cp_cli, name='cp')
dbfs_group.add_command(mv_cli, name='mv')
dbfs_group.add_command(cat_cli, name='cat')



# Databricks CLI
# Copyright 2022 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from databricks_cli.click_types import MetastoreIdClickType, WorkspaceIdClickType, JsonClickType
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.unity_catalog.utils import hide, json_file_help, json_string_help, \
    mc_pretty_format
from databricks_cli.unity_catalog.api import UnityCatalogApi
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, json_cli_base


#################  Metastore Commands  #####################


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Create a metastore.')
@click.option('--name', required=True, help='Name of the new metastore.')
@click.option('--storage-root', required=True,
              help='Storage root URL for the new metastore.')
@click.option('--region', required=False,
              help='Region for the new metastore. ' +
                   'Only workspaces located in the same region can be assigned to this metastore.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def create_metastore_cli(api_client, name, storage_root, region):
    """
    Create new metastore.
    """
    metastore_json = UnityCatalogApi(api_client).create_metastore(name, storage_root, region)
    click.echo(mc_pretty_format(metastore_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List metastores.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_metastores_cli(api_client):
    """
    List metastores.
    """
    metastores_json = UnityCatalogApi(api_client).list_metastores()
    click.echo(mc_pretty_format(metastores_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Get a metastore.')
@click.option('--id', 'metastore_id', required=True, type=MetastoreIdClickType(),
              help='Unique identifier of the metastore to get.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_metastore_cli(api_client, metastore_id):
    """
    Get a metastore.
    """
    metastore_json = UnityCatalogApi(api_client).get_metastore(metastore_id)
    click.echo(mc_pretty_format(metastore_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Update a metastore.')
@click.option('--id', 'metastore_id', required=True, type=MetastoreIdClickType(),
              help='Unique identifier of the metastore to update.')
@click.option('--new-name', default=None, help='New name of the metastore.')
@click.option('--storage-root-credential-id', default=None,
              help='Storage Credential ID to access storage root.')
@click.option('--delta-sharing-scope', default=None,
              help='Delta sharing scope. Can be INTERNAL or INTERNAL_AND_EXTERNAL.')
@click.option('--delta-sharing-recipient-token-lifetime-in-seconds', default=None, type=int,
              help=(
                'The token lifetime determines how long a generated Delta Sharing token is '
                'valid for. 0 for no expiration.'))
@click.option('--delta-sharing-organization-name', default=None,
              help=(
                'The organization name of a Delta Sharing entity. '
                'The name will be used in Databricks-to-Databricks Delta Sharing as the official '
                'name.'))
@click.option('--owner', default=None, help='Owner of the metastore.')
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='PATCH', path='/metastores/{id}'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='PATCH', path='/metastores/{id}'))
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def update_metastore_cli(api_client, metastore_id, new_name,
                         storage_root_credential_id,
                         delta_sharing_scope, delta_sharing_recipient_token_lifetime_in_seconds,
                         delta_sharing_organization_name, owner, json_file, json):
    """
    Update a metastore.

    The public specification for the JSON request is in development.
    """
    has_delta_sharing_flag = (
        (delta_sharing_scope is not None) or
        (delta_sharing_recipient_token_lifetime_in_seconds is not None) or
        (delta_sharing_organization_name is not None))
    if ((new_name is not None) or
        (storage_root_credential_id is not None) or has_delta_sharing_flag or (owner is not None)):
        if (json_file is not None) or (json is not None):
            raise ValueError('Cannot specify JSON if any other update flags are specified')
        data = {
            'name': new_name,
            'storage_root_credential_id': storage_root_credential_id,
            'delta_sharing_scope': delta_sharing_scope,
            'delta_sharing_recipient_token_lifetime_in_seconds':
            delta_sharing_recipient_token_lifetime_in_seconds,
            'delta_sharing_organization_name': delta_sharing_organization_name,
            'owner': owner
        }
        metastore_json = UnityCatalogApi(api_client).update_metastore(metastore_id, data)
        click.echo(mc_pretty_format(metastore_json))
    else:
        json_cli_base(json_file, json,
                  lambda json: UnityCatalogApi(api_client).update_metastore(metastore_id, json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Delete a metastore.')
@click.option('--id', 'metastore_id', required=True, type=MetastoreIdClickType(),
              help='Unique identifier of the metastore to delete.')
@click.option('--force', '-f', is_flag=True, default=False)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_metastore_cli(api_client, metastore_id, force):
    """
    Delete a metastore.
    """
    UnityCatalogApi(api_client).delete_metastore(metastore_id, force)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Get summary info of current metastore.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def metastore_summary_cli(api_client):
    """
    Get metastore summary.
    """
    summary_json = UnityCatalogApi(api_client).get_metastore_summary()
    click.echo(mc_pretty_format(summary_json))

@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Get metastore assignment for workspace.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_metastore_assignment_cli(api_client):
    """
    Get current metastore assignment for workspace.
    """
    assign_json = UnityCatalogApi(api_client).get_current_metastore_assignment()
    click.echo(mc_pretty_format(assign_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Assign a metastore to a workspace.')
@click.option('--workspace-id', 'workspace_id', required=True, type=WorkspaceIdClickType(),
              help='Unique identifier of the workspace for the metastore assignment.')
@click.option('--metastore-id', 'metastore_id', required=True, type=MetastoreIdClickType(),
              help='Unique identifier of the metastore to assign to the workspace.')
@click.option('--default-catalog-name', 'default_catalog_name', required=False,
              default='hive_metastore',
              help='Name of the default catalog to use with the metastore ' +
                   '(default: "hive_metastore").')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def assign_metastore_cli(api_client, workspace_id, metastore_id, default_catalog_name):
    """
    Assign a metastore to a specified workspace.

    If the workspace already has a metastore assigned, it is updated.
    """
    resp = UnityCatalogApi(api_client).create_metastore_assignment(workspace_id, metastore_id,
                                                                   default_catalog_name)
    # resp will just be an empty object ('{}') but it's good to print *something*
    click.echo(mc_pretty_format(resp))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Unassigns a metastore from a workspace.')
@click.option('--workspace-id', 'workspace_id', required=True, type=WorkspaceIdClickType(),
              help='Unique identifier of the workspace.')
@click.option('--metastore-id', 'metastore_id', required=True, type=MetastoreIdClickType(),
              help='Unique identifier of the metastore to unassign from the workspace.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def unassign_metastore_cli(api_client, workspace_id, metastore_id):
    """
    Unassign a metastore from a workspace.
    """
    resp = UnityCatalogApi(api_client).delete_metastore_assignment(workspace_id, metastore_id)
    # resp will just be an empty object ('{}') but it's good to print *something*
    click.echo(mc_pretty_format(resp))


@click.group()
def metastores_group():  # pragma: no cover
    pass


def register_metastore_commands(cmd_group):
    # Register deprecated "verb-noun" commands for backward compatibility.
    cmd_group.add_command(hide(create_metastore_cli), name='create-metastore')
    cmd_group.add_command(hide(list_metastores_cli), name='list-metastores')
    cmd_group.add_command(hide(get_metastore_cli), name='get-metastore')
    cmd_group.add_command(hide(update_metastore_cli), name='update-metastore')
    cmd_group.add_command(hide(delete_metastore_cli), name='delete-metastore')
    cmd_group.add_command(hide(metastore_summary_cli), name='metastore-summary')
    cmd_group.add_command(hide(assign_metastore_cli), name='assign-metastore')
    cmd_group.add_command(hide(unassign_metastore_cli), name='unassign-metastore')
    cmd_group.add_command(hide(get_metastore_assignment_cli),
                          name='get-metastore-assignment')

    # Register command group.
    metastores_group.add_command(create_metastore_cli, name='create')
    metastores_group.add_command(list_metastores_cli, name='list')
    metastores_group.add_command(get_metastore_cli, name='get')
    metastores_group.add_command(update_metastore_cli, name='update')
    metastores_group.add_command(delete_metastore_cli, name='delete')
    metastores_group.add_command(metastore_summary_cli, name='get-summary')
    metastores_group.add_command(assign_metastore_cli, name='assign')
    metastores_group.add_command(unassign_metastore_cli, name='unassign')
    metastores_group.add_command(get_metastore_assignment_cli, name='get-assignment')
    cmd_group.add_command(metastores_group, name='metastores')
